require "forwardable"
require "time"
require "totvs/password_vault/parsers/json_parser"

module Totvs
  module PasswordVault
    class Base
      class Error < StandardError; end
      class RequiredParameter < Error; end

      extend Forwardable
      include JsonParser

      def_delegators :connection, :get, :post, :delete

      # @!attribute [w] connection
      #   @return [Connection]
      attr_writer :connection

      def initialize(connection: nil)
        @connection = connection
      end

      def connection
        @connection ||= Connection.new
      end

      # @param id [String] the id key to fetch info from
      # @return [Hash<Symbol, String>] content the saved content
      # @raise [Totvs::PasswordVault::Connection::RequestFailure] if the request was not a success
      def retrieve(id:)
        headers = { "Accept" => "application/json" }
        response = get(path: build_path(id), headers: headers)
        response = parse_json response.body

        result = {}
        response[translated_group].each do |key, value|
          result[(field_key(key) || key).to_sym] = value
        end

        result
      end

      # @param id [String] the id key to save info
      # @param kwargs [Hash]
      def save(id:, **kwargs)
        headers = {
          "Accept" => "application/json",
          "Content-Type" => "application/json"
        }
        validate_params!(kwargs)

        post path: build_path(id), body: save_body(kwargs), headers: headers
      end

      # @param id [String] the id key to be removed
      # @raise [Totvs::PasswordVault::Connection::RequestFailure] if the request was not a success
      def destroy(id:)
        headers = { "Accept" => "application/json" }
        delete path: build_path(id), headers: headers
      end

      def required_params
        fail NotImplementedError
      end

      def data_group
        fail NotImplementedError
      end

      protected

      # validate presence of all required params
      def validate_params!(params)
        required_params.each do |req_param|
          fail RequiredParameter, "Required parameter {#{req_param}}" if params[req_param].nil?
        end
      end

      # mount body to save request
      # @param params [Hash]
      # @return [Hash] translated body params
      def save_body(params)
        body = {}
        params.each do |key, value|
          body[field_translation(key.to_s).to_sym] = prepare_value_for_saving(value) if field_translation(key.to_s)
        end
        body.delete_if { |_, value| value.nil? || value.empty? }
      end

      # treat some type of value objects
      # @param value [Object] to be prepared
      # @return value with some kind of treatment
      def prepare_value_for_saving(value)
        case value
        when Array
          value.join(", ")
        when Time
          value.iso8601
        when Hash
          value.to_json
        else
          value
        end
      end

      def field_translation(key)
        translator.translate_field(data_group, key)
      end

      def field_key(name)
        translator.field_key(data_group, name)
      end

      def translator
        @translator ||= Translator.new
      end

      def translated_group
        translator.translate_group(data_group)
      end

      def build_path(key)
        "/iso/coe/#{translated_group}/#{key}"
      end
    end
  end
end
