require "oauth"

module Totvs
  module PasswordVault
    class Connection
      class Error < StandardError; end
      class TimeoutError < Error; end
      class RegistryNotFoundError < Error; end
      class InactiveRegistryError < Error; end
      class EnvironmentVariableError < Error; end

      class RequestFailure < Error
        attr_reader :response

        # @param message [String]
        # @param response [Net::HTTPResponse]
        def initialize(message = nil, response:)
          super("Status code: #{response.code}. Body: #{response.body}. #{message}")

          @response = response
        end
      end

      # @!attribute [w] consumer_key
      #   @return [String]
      attr_writer :consumer_key

      # @!attribute [w] consumer_secret
      #   @return [String]
      attr_writer :consumer_secret

      # @!attribute [w] vault_base_url
      #   @return [String]
      attr_writer :vault_base_url

      # @!attribute [w] vault_access_token
      #   @return [String]
      attr_writer :vault_access_token

      # @!attribute [w] vault_access_token_secret
      #   @return [String]
      attr_writer :vault_access_token_secret

      # Returns the oauth consumer
      # @return [OAuth::Consumer]
      def consumer
        @consumer ||= OAuth::Consumer.new(
          consumer_key,
          consumer_secret,
          site: vault_base_url
        )
      end

      # Returns the oauth access token
      # @return [OAuth::AccessToken]
      def access_token
        @access_token ||= OAuth::AccessToken.new(
          consumer,
          vault_access_token,
          vault_access_token_secret
        )
      end

      # Alias access token
      # @return [OAuth::AccessToken]
      def requester
        access_token
      end

      # Makes a http request
      #
      # @param http_method [Symbol]
      # @param path [String]
      # @param args [...]
      #
      # @return [Net::HTTPResponse]
      #
      # @raise [TimeoutError]
      def make_request(http_method, path, *args)
        validate_environment_variables!

        requester.request http_method, path, *args
      rescue Timeout::Error => e
        raise TimeoutError, "#{http_method.to_s.upcase}: #{path}. #{e}"
      end

      # @return [Response] response
      # @raise [RequestFailure] if the request was not a success
      # @raise [TimeoutError]
      def get(path:, headers: {})
        response = make_request(:get, path, headers)
        validate_success! response

        wrap_response response
      end

      # @return [Response] response
      # @raise [RequestFailure] if the request was not a success
      # @raise [TimeoutError]
      def post(path:, body: nil, headers: {})
        response = make_request(:post, path, body, headers)
        validate_success! response

        wrap_response response
      end

      # @return [Response] response
      # @raise [RequestFailure] if the request was not a success
      # @raise [TimeoutError]
      def put(path:, body: nil, headers: {})
        response = make_request(:put, path, body, headers)
        validate_success! response

        wrap_response response
      end

      # @return [Response] response
      # @raise [RequestFailure] if the request was not a success
      # @raise [TimeoutError]
      def delete(path:, headers: {})
        response = make_request(:delete, path, headers)
        validate_success! response

        wrap_response response
      end

      def required_environment_variables
        %w(
          VAULT_CONSUMER_KEY
          VAULT_CONSUMER_SECRET
          VAULT_BASE_URL
          VAULT_ACCESS_TOKEN
          VAULT_ACCESS_TOKEN_SECRET
        )
      end

      # @return [String]
      def consumer_key
        @consumer_key ||= ENV["VAULT_CONSUMER_KEY"]
      end

      # @return [String]
      def consumer_secret
        @consumer_secret ||= ENV["VAULT_CONSUMER_SECRET"]
      end

      # @return [String]
      def vault_base_url
        @vault_base_url ||= ENV["VAULT_BASE_URL"]
      end

      # @return [String]
      def vault_access_token
        @vault_access_token ||= ENV["VAULT_ACCESS_TOKEN"]
      end

      # @return [String]
      def vault_access_token_secret
        @vault_access_token_secret ||= ENV["VAULT_ACCESS_TOKEN_SECRET"]
      end

      # Wraps the response to a Response object
      # @return [Response]
      def wrap_response(response)
        Response.new status_code: response.code.to_i,
                     body: response.body,
                     content_type: response.content_type
      end

      protected

      # validates presence of environment variables
      def validate_environment_variables!
        required_environment_variables.each do |variable|
          fail EnvironmentVariableError, "Missing environment variable: #{variable}" unless ENV[variable]
        end
      end

      def validate_success!(response)
        if response.is_a? Net::HTTPNotFound
          fail RegistryNotFoundError
        elsif response.is_a? Net::HTTPGone
          fail InactiveRegistryError, "This registry is inactive and can't be updated"
        elsif response.is_a? Net::HTTPSuccess
          return response
        else
          message = "Response class: #{response}"
          fail RequestFailure.new message, response: response
        end
      end
    end
  end
end
